import os

import pytest


@pytest.fixture(autouse=True, name="settings_module", scope="session")
def settings_module() -> None:
    os.environ["DJANGO_SETTINGS_MODULE"] = "tests.settings"


@pytest.fixture(autouse=True, name="settings", scope="session")
def _settings(settings_module):
    import django
    from django.conf import settings

    django.setup()

    return settings


@pytest.fixture(scope="function", autouse=True)
def _dj_autoclear_mailbox() -> None:
    # Override the `_dj_autoclear_mailbox` test fixture in `pytest_django`.
    # This works around https://github.com/pytest-dev/pytest-django/issues/993
    pass



from unittest.mock import patch

import pytest
from django.http import HttpRequest, HttpResponse
from django.test import AsyncRequestFactory, RequestFactory

from django_svcs import middleware


@pytest.fixture(name="svcs_from")
def mock_svcs_from():
    with patch("django_svcs.middleware.svcs_from") as mock:
        yield mock


@pytest.fixture(name="sync_view")
def _sync_view():
    def view(request: HttpRequest) -> HttpResponse:
        return HttpResponse(b"Sync response")

    return view


@pytest.fixture(name="async_view")
def _async_view():
    async def view(request: HttpRequest) -> HttpResponse:
        return HttpResponse(b"Async response")

    return view


@pytest.mark.django_db
def test_sync_middleware(svcs_from, sync_view) -> None:
    request = RequestFactory().get("/")
    chain = middleware.request_container(sync_view)
    response = chain(request)

    assert response.content.decode() == "Sync response"
    assert svcs_from.called
    svcs_from.assert_called_with(request)


@pytest.mark.django_db
@pytest.mark.asyncio
async def test_async_middleware(svcs_from, async_view) -> None:
    request = AsyncRequestFactory().get("/")
    chain = middleware.request_container(async_view)
    response = await chain(request)

    assert response.content.decode() == "Async response"
    assert svcs_from.called
    svcs_from.assert_called_with(request)



from .apps import close_registry, get_registry, svcs_from

__all__ = ["close_registry", "get_registry", "svcs_from"]

try:
    from ._version import __version__, __version_tuple__  # type: ignore
except ImportError:
    __version__ = "0.0.0"
    __version_tuple__ = (0, 0, 0)



import djp


@djp.hookimpl
def installed_apps():
    return ["django_svcs"]


@djp.hookimpl
def urlpatterns():
    return []


@djp.hookimpl
def settings(current_settings) -> None:
    # find django_svcs in INSTALLED_APPS and move it so it's right after django.contrib.sessions
    installed_apps = current_settings["INSTALLED_APPS"]
    try:
        insert_index = installed_apps.index("django.contrib.sessions")
    except ValueError:
        insert_index = None

    if insert_index is not None:
        installed_apps.remove("django_svcs")

        installed_apps.insert(insert_index + 1, "django_svcs")


@djp.hookimpl
def middleware():
    return [
        djp.After("django.contrib.sessions.middleware.SessionMiddleware"),
        "django_svcs.middleware.request_container",
    ]



import asyncio
from typing import Callable

from django.http import HttpRequest, HttpResponse
from django.utils.decorators import sync_and_async_middleware

from .apps import svcs_from

# pyright: reportRedeclaration=false


@sync_and_async_middleware
def request_container(get_response: Callable):
    """Middleware that attaches a service container to the request"""
    if asyncio.iscoroutinefunction(get_response):

        async def middleware(request: HttpRequest) -> HttpResponse:
            """
            Asynchronous middleware function which attaches a
            service container to the request.

            """
            with svcs_from(request):
                response = await get_response(request)
                return response

    else:

        def middleware(request: HttpRequest) -> HttpResponse:
            """
            Synchronous middleware function which attaches a
            service container to the request.

            """
            with svcs_from(request):
                response = get_response(request)
                return response

    return middleware



import pytest
from django.core.exceptions import ImproperlyConfigured

from django_svcs import apps


def test_requires_middleware(settings) -> None:
    settings.MIDDLEWARE = []
    with pytest.raises(ImproperlyConfigured):
        apps.SvcsConfig("django_svcs", apps).ready()


def test_get_registry() -> None:
    assert apps.get_registry() is not None



from importlib import import_module
from typing import Protocol, cast

import svcs
from asgiref.local import Local
from django.apps import AppConfig, apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest

_KEY_CONTAINER = "svcs_container"
_NON_REQUEST_CONTEXT = Local()


class DjangoRegistryContainer(Protocol):
    registry: svcs.Registry


class SvcsConfig(AppConfig):
    """
    Django AppConfig for the services module

    This appconfig creates a registry for services and initializes it with services from the project
    and all installed apps that provide `services.svcs_init` functions.

    In order to be fully functional, it's necessary to add request wrapping middleware to your
    project by including the `django_svcs.middleware.request_container` middleware in your
    `MIDDLEWARE` setting. It must be before any middleware that accesses services.

    """

    name = "django_svcs"
    verbose_name = "Services"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def ready(self) -> None:
        registry = svcs.Registry()

        if "django_svcs.middleware.request_container" not in settings.MIDDLEWARE:
            raise ImproperlyConfigured(
                "The request_container middleware must be included in your MIDDLEWARE setting"
            )

        for app_config in apps.get_app_configs():
            try:
                services_for_app = import_module(f"{app_config.name}.services")
            except ImportError:
                continue

            if hasattr(services_for_app, "svcs_init"):
                services_for_app.svcs_init(registry)

        self.registry = registry


def get_registry(app_config: DjangoRegistryContainer | None = None) -> svcs.Registry:
    """Get the registry from the app config, or from the registered appconfig"""
    registry_holder = (
        cast(DjangoRegistryContainer, apps.get_app_config("django_svcs"))
        if app_config is None
        else app_config
    )
    return registry_holder.registry


def close_registry(app_config: DjangoRegistryContainer | None = None) -> None:
    """Close the registry from the app config, or from the registered appconfig"""
    ...


def svcs_from(
    request: HttpRequest | None = None,
) -> svcs.Container:
    """
    Get the current container from the request, if provided, or from a thread/task-local storage
    if no request is provided.

    """
    context = request or _NON_REQUEST_CONTEXT
    if not hasattr(context, _KEY_CONTAINER):
        setattr(context, _KEY_CONTAINER, svcs.Container(registry=get_registry()))

    return getattr(context, _KEY_CONTAINER)
